import numpy as np

from .base import Distribution
from ..lib.utils import ArrayLike
from ..lib.decorator import adjust_dims
from ..math.mittagleffler import mlf


class ColeCole(Distribution):
    """
    Functions based on Cole-Cole distribution
    """
    name = 'Cole-Cole'
    parameter = [r'\alpha']
    bounds = [(0, 1)]

    @staticmethod
    def distribution(tau, tau0, alpha):
        """
        Distribution of correlation times

        .. math::
            G(\ln\\tau) = \\frac{1}{2\pi} \\frac{\\sin(\\pi\\alpha)}{\cosh[\\alpha\ln(\\tau/\\tau_0)] + \cos(\\alpha \pi))}

        Args:
            tau:
            tau0:
            alpha:
        """
        z = np.log(tau/tau0)
        return np.sin(np.pi*alpha)/(np.cosh(z*alpha) + np.cos(alpha*np.pi))/(2*np.pi)

    @staticmethod
    @adjust_dims
    def susceptibility(omega: ArrayLike, tau: ArrayLike, alpha: float) -> ArrayLike:
        r"""
        Complex susceptibility

        .. math::
            \chi(\omega, \tau, \alpha) = \frac{1}{[1-(i\omega\tau)^\alpha]}

        Args:
            omega (array-like): Frequency axis in 1/s (not Hz).
            tau (array-like): Correlation times :math:`\tau_\text{CC}` in s.
            alpha (float): Shape parameter.

        """
        val = 1/(1+(1j*omega*tau)**alpha)
        return np.conjugate(val)

    @staticmethod
    @adjust_dims
    def specdens(omega, tau, alpha):
        """
        Spectral density

        .. math::
            J(\omega, \\tau, \\alpha) = \\frac{\sin(\\alpha\pi/2)(\omega\\tau)^\\alpha}{\omega[1+(\omega\\tau)^{2\\alpha} + 2\\pi\cos(\\alpha\pi/2)(\omega\\tau)^\\alpha]}

        Args:
            omega:
            tau:
            alpha:
        """
        omtau = (omega*tau)**alpha
        return np.sin(alpha*np.pi/2) * omtau / (1 + omtau**2 + 2*np.cos(alpha*np.pi/2)*omtau) / omega

    @staticmethod
    @adjust_dims
    def correlation(t, tau0, alpha):
        """
        Correlation function :math:`C(t)`

        .. math::
            C(t) = E_\\alpha[-(t/\\tau_0)^\\alpha]

        with Mittag-Leffler function :math:`E_a(x)`

        Args:
            t:
            tau0:
            alpha:
        """
        return mlf(-(t/tau0)**alpha, alpha)
