import inspect
from typing import Sized

from numpy import inf

from ._meta import MultiModel
from .parameter import Parameters


class Model(object):
    def __init__(self, model, *args, **kwargs):
        self.idx = kwargs.pop('idx', None)

        self.is_multi = False
        if inspect.isclass(model) or isinstance(model, MultiModel):
            self._init_from_class(model)
        elif inspect.isfunction(model):
            self._init_from_function(model)
        else:
            raise ValueError(f'No idea how to use datatype {model!r}.')

        self.lb = [i if i is not None else -inf for i in self.lb]
        self.ub = [i if i is not None else inf for i in self.ub]

        self.parameter = Parameters()
        self.global_parameter = {}
        self.is_complex = None
        self._complex_part = False

        if 'complex' in kwargs:
            self.set_complex(kwargs.pop('complex'))

        if args:
            self.fun_args = args
        else:
            self.fun_args = []

        self.fun_kwargs.update(kwargs)

    def _init_from_function(self, func):
        self.name = str(func)
        self._int_func = func
        self._int_iter = func

        self.params = []
        self.fun_kwargs = {}
        for k, v in inspect.signature(func).parameters.items():
            if v.default != inspect.Parameter.empty:
                self.fun_kwargs[k] = v.default
            else:
                self.params.append(k)
        # first parameter is x
        self.params.pop(0)

        self.lb = [None] * len(self.params)
        self.ub = [None] * len(self.params)

    def _init_from_class(self, model):
        self.name = model.name
        self.params = model.params
        self._int_func = model.func
        if hasattr(model, 'subs'):
            self._int_iter = model.subs
            self.is_multi = True
        else:
            self._int_iter = model.func


        try:
            self.lb, self.ub = list(zip(*model.bounds))
        except AttributeError:
            self.lb = [None] * len(self.params)
            self.ub = [None] * len(self.params)

        if isinstance(model, MultiModel):
            self.fun_kwargs = model.fun_kwargs
        else:
            self.fun_kwargs = {k: v.default for k, v in inspect.signature(model.func).parameters.items()
                               if v.default is not inspect.Parameter.empty}

    def set_complex(self, state):
        if state not in [None, 'complex', 'real', 'imag']:
            raise ValueError('"complex" argument is not None, "complex", "real", "imag"')

        self.is_complex = state
        if state in ['real', 'imag']:
            self._complex_part = state
        else:
            self._complex_part = False

    def set_global_parameter(self, idx, p, var=None, lb=None, ub=None, default_bounds=False):
        if idx is None:
            self.parameter = Parameters()
            self.global_parameter = {}
            return

        if default_bounds:
            if lb is None:
                lb = [self.lb[i] for i in idx]
            if ub is None:
                ub = [self.lb[i] for i in idx]

        gp = self.parameter.add_parameter(p, var=var, lb=lb, ub=ub)
        for k, v in zip(idx, gp):
            self.global_parameter[k] = v

        return gp

    @staticmethod
    def _prep(param_len, val):
        if isinstance(val, Sized):
            if len(val) != param_len:
                raise ValueError('Length mismatch for global parameter')
        else:
            return [val] * param_len

    def __str__(self):
        msg = 'Model: ' + self.name
        return msg

    def __len__(self):
        return len(self.params)

    def func(self, p, x, **kwargs):
        if not kwargs:
            kwargs = self.fun_kwargs

        f = self._int_func(x, *p, *self.fun_args, **kwargs)

        if self._complex_part:
            if self._complex_part == 'real':
                return f.real
            else:
                return f.imag

        return f

    def sub(self, p, x, **kwargs):
        if not self.is_multi:
            return [self.func(p, x, **kwargs)]

        else:
            if not kwargs:
                kwargs = self.fun_kwargs

            if self._complex_part:
                if self._complex_part == 'real':
                    return [f.real for f in self._int_iter(x, *p, *self.fun_args, **kwargs)]
                else:
                    return [f.imag for f in self._int_iter(x, *p, *self.fun_args, **kwargs)]

            return list(self._int_iter(x, *p, *self.fun_args, **kwargs))
